/*
 * pcie_template_device.c
 *
 * Copyright (c) 2010 Adnan Khaleel <akhaleel at cray com>
 *                    Cray Inc
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* This is essentially an example for a device that utilises PCIe and MSI-X */

#include "hw.h"
#include "pc.h"
#include "pci_ids.h"
#include "msix.h"
#include "pcie.h"

#include "pcie_msix_template.h"

/* Debug code */
#define DEBUG_PCIE_MSIX_TEMPLATE

#ifdef DEBUG_PCIE_MSIX_TEMPLATE
#define PRINT_DEBUG(format, ...) printf(format, __VA_ARGS__);
#define PRINT_DEBUG0(format) printf(format);
#else
#define PRINT_DEBUG(format, ...)
#define PRINT_DEBUG0(format)
#endif

#define DEFINE_DEVICE_PROPERTIES(_state, _conf)

#define PCIE_MSIX_VID                   PCI_VENDOR_ID_CRAY
#define PCIE_MSIX_DID                   PCI_DEVICE_ID_CRAY_ASIC
#define PCIE_MSIX_SS_DID                0x100
#define PCIE_MSIX_VERSION               0x1
#define PCIE_MSIX_MSI_SUPPORTED_FLAGS   PCI_MSI_FLAGS_64BIT
#define PCIE_MSIX_MSI_NR_VECTOR         1        
#define PCIE_MSIX_MSI_OFFSET            0x70
#define PCIE_MSIX_SSVID_OFFSET          0x80
#define PCIE_MSIX_SVID                  0
#define PCIE_MSIX_SSID                  0
#define PCIE_MSIX_EXP_OFFSET            0x90
#define PCIE_MSIX_AER_OFFSET            0x100

/* PCI BUS commands */
#define CMD_IOREAD	        0x02
#define CMD_IOWRITE	        0x03
#define CMD_MEMREAD	        0x06
#define CMD_MEMWRITE	    0x07
#define CMD_CONFIGREAD	    0x0A
#define CMD_CONFIGWRITE	    0x0B

static const unsigned long long BAR_Regions[6][2] = 
{
    // len , type 
    { 0x2000000000ull, PCI_BASE_ADDRESS_SPACE_MEMORY} ,  //BAR0
    { 0x2000000ull,    PCI_BASE_ADDRESS_SPACE_IO    } ,  //BAR1
    { 0, 0} ,  //BAR2
    { 0, 0} ,  //BAR3
    { 0, 0} ,  //BAR4
    { 0, 0}    //BAR5 
};

typedef struct PCIE_MSIX_DEVState_St {
    PCIDevice dev;
    int mmio_index;
} PCIE_MSIX_DEVState;


/*
static void 
pcie_msix_irq(void)
{
    PRINT_DEBUG("\n%s: low_irq\n", __FUNCTION__);
    // How do you raise an interrupt?

}
*/

static void 
pcie_msix_io_write(PCIDevice *d, uint32_t addr, uint32_t val, int len)
{
    uint32_t reg_base, reg_size;
    int bar_region=0, i;
    
    // Is this the correct?
    for(i=0; i<PCI_NUM_REGIONS; i++) 
    {
        reg_base = d->io_regions[i].addr;

        if (reg_base < 0) 
            continue; /*-1 means not mapped */

        reg_size = d->io_regions[i].size;
        if((reg_base <= addr) && (reg_base+reg_size > addr)) 
        {
            bar_region = i;
            break;
        }
    }
    
    PRINT_DEBUG("\n%s: reg_base: %08X\n", __FUNCTION__, reg_base);
    PRINT_DEBUG("\n%s: io_write: reg_base=0x%x len=0x%x addr=0x%x val=0x%x bar=%d\n", __FUNCTION__, reg_base, len, addr, val, bar_region);  
    // Do Write
}

static uint32_t
pcie_msix_io_read(PCIDevice *d, uint32_t addr, int len)
{
    uint32_t reg_base, reg_size, val = 0;
    int bar_region=0, i;

    // Is this the correct?
    for(i=0; i<PCI_NUM_REGIONS; i++) 
    {
        reg_base = d->io_regions[i].addr;

        if (reg_base < 0) 
            continue; /*-1 means not mapped */

        reg_size = d->io_regions[i].size;
        if((reg_base <= addr) && (reg_base+reg_size > addr)) 
        {
            bar_region = i;
            break;
        }
    }

    PRINT_DEBUG("\n%s: io_read %d bytes. Addr=0x%x bar=%d",__FUNCTION__, len, addr, bar_region);
    // Do Read
    PRINT_DEBUG("\n%s: io_read %d bytes. Addr=0x%x val=0x%x bar=%d",__FUNCTION__, len, addr, val, bar_region);
    return val;
}

static void
pcie_msix_io_writeb(void* opaque, uint32_t addr, uint32_t val)
{
    pcie_msix_io_write(opaque, addr, val, 1);
}

static void
pcie_msix_io_writew(void* opaque, uint32_t addr, uint32_t val)
{
    pcie_msix_io_write(opaque, addr, val, 2);
}

static void
pcie_msix_io_writel(void* opaque, uint32_t addr, uint32_t val)
{
    pcie_msix_io_write(opaque, addr, val, 4);
}

static uint32_t
pcie_msix_io_readb(void* opaque, uint32_t addr)
{
    return pcie_msix_io_read(opaque, addr, 1);
}

static uint32_t
pcie_msix_io_readw(void* opaque, uint32_t addr)
{
    return pcie_msix_io_read(opaque, addr, 2);
}

static uint32_t
pcie_msix_io_readl(void* opaque, uint32_t addr)
{
    return pcie_msix_io_read(opaque, addr, 4);
}

static void 
pcie_msix_mem_write(void *opaque, uint32_t addr, uint32_t val, int len)
{
    uint32_t reg_base, reg_size;
    uint32_t base_addr;
    int bar_region=0, i;
    PCIDevice *d=opaque;

    PRINT_DEBUG("\n%s: mem_write: addr: 0x%x\n", __FUNCTION__, addr);

    for(i=0; i<PCI_NUM_REGIONS; i++) 
    {
        reg_base = d->io_regions[i].addr;
        if (reg_base < 0) 
            continue; /*-1 means not mapped */

        reg_size = d->io_regions[i].size;
        if((reg_base <= addr) && (reg_base+reg_size > addr)) 
        {
            bar_region = i;
            break;
        }
    }
    base_addr = d->io_regions[bar_region].addr;
    PRINT_DEBUG("\n%s: mem_write BAR Base address=0x%x Addr=0x%x val=0x%x bar=%d\n", __FUNCTION__, base_addr, addr, val,bar_region);
    PRINT_DEBUG("\n%s: mem_write REAL address: 0x%x\n", __FUNCTION__, base_addr + addr);

    PRINT_DEBUG("%s: PCIIOREGIONS:\n", __FUNCTION__);

    for(i=0;i< PCI_NUM_REGIONS-1;i++) 
    {
        PRINT_DEBUG("\t%d Addr: 0x%"FMT_PCIBUS" Size: 0x%"FMT_PCIBUS". Type: 0x%x\n", i, d->io_regions[i].addr, d->io_regions[i].size, d->io_regions[i].type);
    }
    // Do Write 
    // This BAR is memory mapped, we perform a DMI access instead of a standard access.
}

static uint32_t
pcie_msix_mem_read(void *opaque, uint32_t addr, int len)
{
    uint32_t val = 0, reg_base, reg_size;
    int bar_region, i;
    PCIDevice *d=opaque;

    PRINT_DEBUG("\n%s: device name: %s\t devfn: %d\n", __FUNCTION__, d->name, d->devfn);

    bar_region = 0;
    for(i=0; i<PCI_NUM_REGIONS; i++) 
    {
        reg_base = d->io_regions[i].addr;
        if (reg_base < 0) continue; 

        reg_size = d->io_regions[i].size;
        if((reg_base <= addr) && (reg_base+reg_size > addr)) 
        {
            bar_region = i;
            break;
        }
    }

    PRINT_DEBUG("\n%s: device name: %s\t devfn: %d\n", __FUNCTION__, d->name, d->devfn);
    PRINT_DEBUG("\n%s: mem_read Addr=0x%x bar=%d\n", __FUNCTION__, addr-reg_base, bar_region);

    // This BAR is memory mapped, we perform a DMI access instead of a standard access.
    
    PRINT_DEBUG("\n%s: io_read %d bytes. Addr=0x%x val=0x%x bar=%d",__FUNCTION__, len, addr, val, bar_region);
    // Do Read and return val
    return val;
}

static void
pcie_msix_mem_writeb(void *opaque, target_phys_addr_t addr, uint32_t val)
{
    pcie_msix_mem_write(opaque, addr, val, 1);
}

static void
pcie_msix_mem_writew(void *opaque, target_phys_addr_t addr, uint32_t val)
{
    pcie_msix_mem_write(opaque, addr, val, 2);
}

static void
pcie_msix_mem_writel(void *opaque, target_phys_addr_t addr, uint32_t val)
{
    pcie_msix_mem_write(opaque, addr, val, 4);
}

static uint32_t 
pcie_msix_mem_readb(void *opaque, target_phys_addr_t addr)
{
    return pcie_msix_mem_read(opaque, addr, 1);
}

static uint32_t 
pcie_msix_mem_readw(void *opaque, target_phys_addr_t addr)
{
    return pcie_msix_mem_read(opaque, addr, 2);
}

static uint32_t 
pcie_msix_mem_readl(void *opaque, target_phys_addr_t addr)
{
    return pcie_msix_mem_read(opaque, addr, 4);
}

static CPUWriteMemoryFunc * const pcie_msix_mem_write_fn[] = {
    pcie_msix_mem_writeb, pcie_msix_mem_writew, pcie_msix_mem_writel
};

static CPUReadMemoryFunc * const pcie_msix_mem_read_fn[] = {
    pcie_msix_mem_readb, pcie_msix_mem_readw, pcie_msix_mem_readl
};

static void 
pcie_msix_io_map(PCIDevice *dev, int region_num, pcibus_t addr, 
                 pcibus_t size, int type)
{
    register_ioport_write(addr, 16, 1, pcie_msix_io_writeb, dev);
    register_ioport_write(addr, 16, 2, pcie_msix_io_writew, dev);
    register_ioport_write(addr, 16, 4, pcie_msix_io_writel, dev);
    register_ioport_read(addr , 16, 1, pcie_msix_io_readb, dev);
    register_ioport_read(addr , 16, 2, pcie_msix_io_readw, dev);
    register_ioport_read(addr , 16, 4, pcie_msix_io_readl, dev);
}

static void 
pcie_msix_mem_map(PCIDevice *dev, int region_num, 
                  pcibus_t addr, pcibus_t size, int type)
{
    PCIE_MSIX_DEVState *d = DO_UPCAST(PCIE_MSIX_DEVState, dev, dev);
    cpu_register_physical_memory(addr, size, d->mmio_index);
}

static void pcie_msix_write_config(PCIDevice *d,
                                   uint32_t address, uint32_t val, int len)
{
    pcie_cap_deverr_write_config(d, address, val, len);
    pcie_cap_flr_write_config(d, address, val, len);
    msix_write_config(d, address, val, len);
    pcie_aer_write_config_vbridge(d, address, val, len);
    pcie_aer_write_config(d, address, val, len);
}

static void pcie_msix_reset(DeviceState *qdev)
{
    PCIDevice *d = DO_UPCAST(PCIDevice, qdev, qdev);
    msix_reset(d);
    pci_bridge_reset(qdev);
    pcie_cap_deverr_reset(d);
}

static void pcie_msix_flr(PCIDevice *d)
{
    pci_device_reset(d);
}

static int pcie_msix_initfn(PCIDevice *pci_dev)
{
    PCIE_MSIX_DEVState *d = DO_UPCAST(PCIE_MSIX_DEVState, dev, pci_dev);
    int rc, i;

    PRINT_DEBUG("%s: PCIE MSIX Device init...\n", __FUNCTION__);

    pci_config_set_vendor_id(d->dev.config, PCIE_MSIX_VID);
    pci_config_set_device_id(d->dev.config, PCIE_MSIX_DID);
    d->dev.config[PCI_REVISION_ID] = PCIE_MSIX_VERSION;
    d->dev.config[PCI_SUBSYSTEM_VENDOR_ID]   = PCIE_MSIX_VID & 0xff;
    d->dev.config[PCI_SUBSYSTEM_VENDOR_ID+1] = (PCIE_MSIX_VID >> 8) & 0xff;
    d->dev.config[PCI_SUBSYSTEM_ID]   = PCIE_MSIX_SS_DID & 0xff;
    d->dev.config[PCI_SUBSYSTEM_ID+1] = (PCIE_MSIX_SS_DID >> 8) & 0xff;

    d->mmio_index = cpu_register_io_memory(pcie_msix_mem_read_fn, pcie_msix_mem_write_fn, d);
    
    for(i=0; i < PCI_NUM_REGIONS -1; i++) { //-1 for the Exp ROM BAR
        if(BAR_Regions[i][0] != 0)
        { 
            if(BAR_Regions[i][1] == PCI_BASE_ADDRESS_SPACE_IO) 
            {
                //io region
                PRINT_DEBUG("%s: Registering Bar %i as I/O BAR\n", __FUNCTION__, i);
                pci_register_bar((PCIDevice *)d, i, BAR_Regions[i][0], PCI_BASE_ADDRESS_SPACE_IO, pcie_msix_io_map);
            } else { 
                //mem region
                PRINT_DEBUG("%s: Registering Bar %i as MEM BAR\n", __FUNCTION__, i);
                pci_register_bar((PCIDevice *)d, i, BAR_Regions[i][0], PCI_BASE_ADDRESS_SPACE_MEMORY, pcie_msix_mem_map);
            }
        }
    }
    d->dev.config[PCI_INTERRUPT_PIN] = 1;

    rc = msix_init(&d->dev, PCIE_MSIX_MSI_NR_VECTOR, 1, 0);
    if (rc < 0) {
        return rc;
    }

    // How do I initialise the pcie capabilities? 
    //rc = pci_pcie_cap_init(d->dev, PCIE_MSIX_EXP_OFFSET, PCI_EXP_TYPE_ENDPOINT, p->port);
    if (rc < 0) {
        return rc;
    }

    pcie_cap_flr_init(&d->dev, &pcie_msix_flr);
    pcie_cap_deverr_init(&d->dev);
    rc = pcie_aer_init(&d->dev, PCIE_MSIX_AER_OFFSET);
    if (rc < 0) {
        return rc;
    }

    PRINT_DEBUG("%s: Init done\n", __FUNCTION__);
    return 0;
}

static int pcie_msix_exitfn(PCIDevice *pci_dev)
{
    PCIE_MSIX_DEVState *d = DO_UPCAST(PCIE_MSIX_DEVState, dev, pci_dev);

    pcie_aer_exit(&d->dev);
    msix_uninit(&d->dev);
    return pci_pcie_cap_exit(&d->dev);
}

void pcie_msix_init(PCIBus *bus)
{
    // Is this how we should be doing this?
    pci_create_simple(bus, -1, "SCLink");
}

static const VMStateDescription vmstate_pcie_msix = {
    .name = "pcie-msix-device",
    .version_id = 1,
    .minimum_version_id = 1,
    .minimum_version_id_old = 1,
    .fields = (VMStateField[]) {
        VMSTATE_PCIE_DEVICE(dev, PCIE_MSIX_DEVState),
        VMSTATE_STRUCT(dev.aer_log, PCIE_MSIX_DEVState, 0, vmstate_pcie_aer_log, struct pcie_aer_log),
        VMSTATE_END_OF_LIST()
    }
};

static PCIDeviceInfo pcie_msix_info = {
    .qdev.name = PCIE_MSIX_DEVICE,
    .qdev.desc = "PCIE MSIX device template",
    .qdev.size = sizeof(PCIE_MSIX_DEVState),
    .qdev.reset = pcie_msix_reset,
    .qdev.vmsd = &vmstate_pcie_msix,
    .is_express = 1,
    .config_write = pcie_msix_write_config,
    .init = pcie_msix_initfn,
    .exit = pcie_msix_exitfn,
    .qdev.props = (Property[]) {        
        DEFINE_PROP_END_OF_LIST(),
    }
};

static void pcie_msix_register(void)
{
    // Where do I register this device?
    //pci_bridge_qdev_register(&pcie_msix_info);
    pci_qdev_register(&pcie_msix_info);
}

device_init(pcie_msix_register);
